__all__ = [
    '_BaseFideliusRepo',
    '_BaseFideliusAdminRepo',
]
from .interface import *

from fidelius.structs import *
import logging
log = logging.getLogger(__file__)


class _BaseFideliusRepo(IFideliusRepo, abc.ABC):
    """Covers a lot of basic functionality common across most storage back-ends.
    """

    _APP_PATH_FORMAT = '/fidelius/{group}/{env}/apps/{app}/{name}'
    _SHARED_PATH_FORMAT = '/fidelius/{group}/{env}/shared/{folder}/{name}'

    _EXPRESSION_APP_FORMAT = '${{__FID__:{name}}}'
    _EXPRESSION_SHARED_FORMAT = '${{__FID__:{folder}:{name}}}'

    _EXPRESSION_PATTERN = re.compile(r'\${__FID__:(?:(?P<folder>\w+):)?(?P<name>[\w/-]+)}')

    def __init__(self, app_props: FideliusAppProps, **kwargs):
        log.debug('_BaseFideliusRepo.__init__')
        self._app_props = app_props

        # Any kwargs should have been handled by implementation specific stuff,
        # unless there's a derpy config so let's warn just in case!
        if kwargs:
            for k, v in kwargs.items():
                log.warning(f'AbstractFideliusRepo for some unhandled kwargs: {k}={v}')

    @property
    def app_props(self) -> FideliusAppProps:
        """The current application properties.
        """
        return self._app_props

    def make_app_path(self, env: Optional[str] = None) -> str:
        """The full path to application specific parameters/secrets.
        """
        return self._APP_PATH_FORMAT.format(group=self.app_props.group,
                                            env=env or self.app_props.env,
                                            app=self.app_props.app,
                                            name='{name}')

    def make_shared_path(self, folder: str, env: Optional[str] = None) -> str:
        """The full path to group shared parameters/secrets.
        """
        return self._SHARED_PATH_FORMAT.format(group=self.app_props.group,
                                            env=env or self.app_props.env,
                                            folder=folder,
                                            name='{name}')

    def get_expression_string(self, name: str, folder: Optional[str] = None) -> str:
        """Return a Fidelius expression string (e.g. to use in configuration
        files) which the `replace` method can parse and replace with
        parameters/secrets fetched from the storage backend.

        The default format of these expressions is:
        - `${__FID__:PARAM_NAME}` for app params/secrets
        - `${__FID__:FOLDER:PARAM_NAME}` for shared params/secrets in the given FOLDER

        :param name: The name of the parameter/secret to create an expression for.
        :param folder: Optional name of shared parameters/secrets to use. This
                       is optional and only applies to shared
                       parameters/secrets. Leaving it blank (default) will
                       return the app specific expression.
        :return: The Fidelius expression string.
        """
        if folder:
            return self._EXPRESSION_SHARED_FORMAT.format(name=name, folder=folder)
        else:
            return self._EXPRESSION_APP_FORMAT.format(name=name)

    def get_full_path(self, name: str, folder: Optional[str] = None, env: Optional[str] = None) -> str:
        """Gets the full path to a parameter/secret.

        Parameter/secret paths can be either application specific (use the
        `app_path`) or shared (use the `shared_path`) and this method should
        determine whether to use a shared or app path based on whether or
        not a folder name was given.

        :param name: The name of the parameter/secret to build a path to.
        :param folder: Optional name of shared parameters/secrets to use. This
                       is optional and only applies to shared
                       parameters/secrets. Leaving it blank (default) will
                       return the app specific parameter/secret path.
        :param env: Optional env value override. Defaults to None, which uses
                    the env declaration from the current app properties.
        :return: The full path to a parameter/secret.
        """
        if folder:
            return self.make_shared_path(folder, env=env).format(name=name)
        else:
            return self.make_app_path(env=env).format(name=name)

    def get(self, name: str, folder: Optional[str] = None, no_default: bool = False) -> Optional[str]:
        """Gets the given parameter/secret from the storage this repo uses.

        Parameters/secrets can be either application specific (use the
        `app_path`) or shared (use the `shared_path`) and this method should
        determine whether to get a shared or app parameter based on whether or
        not a folder name was given.

        Unless disabled by the `no_default` parameter, this method will attempt
        to fetch a parameter using the `env=default` if one was not found for
        the env in the current app properties.

        :param name: The name of the parameter/secret to get.
        :param folder: Optional name of shared parameters/secrets to use. This
                       is optional and only applies to shared
                       parameters/secrets. Leaving it blank (default) will
                       return the app specific parameter/secret.
        :param no_default: If True, does not try and get the default value if no
                           value was found for the current set environment.
        :return: The requested parameter/secret or None if it was not found.
        """
        log.debug('_BaseFideliusRepo.get(name=%s, folder=%s, no_default=%s))', name, folder, no_default)
        if folder:
            val = self.get_shared_param(name=name, folder=folder)
            log.debug('_BaseFideliusRepo.get->get_shared_param val=%s', val)
            if val is not None:
                return val

            if no_default:
                log.debug('_BaseFideliusRepo.get->(shared) no_default STOP!')
                return None

            log.debug('_BaseFideliusRepo.get->(shared) Lets try the default!!!')
            return self.get_shared_param(name=name, folder=folder, env='default')
        else:
            val = self.get_app_param(name=name)
            log.debug('_BaseFideliusRepo.get->get_app_param val=%s', val)
            if val is not None:
                return val

            if no_default:
                log.debug('_BaseFideliusRepo.get->(app) no_default STOP!')
                return None

            log.debug('_BaseFideliusRepo.get->(app) Lets try the default!!!')
            return self.get_app_param(name=name, env='default')

    def replace(self, string: str, no_default: bool = False) -> str:
        """Take in a containing a Fidelius parameter/secret configuration
        expression (e.g. read from a config file), parses it, fetches the
        relevant parameter/secret and returns.

        The default format of these expressions is:
        - `${__FID__:PARAM_NAME}` for app params/secrets
        - `${__FID__:FOLDER:PARAM_NAME}` for shared params/secrets in the given FOLDER

        An empty string is returned if the parameter was not found and if the
        string does not match the expression format, it will be returned
        unchanged.

        :param string: The expression to replace with an actual parameter/secret
        :param no_default: If True, does not try and get the default value if no
                           value was found for the current set environment.
        :return: The requested value, an empty string or the original string
        """
        m = self._EXPRESSION_PATTERN.match(string)
        if m:
            return self.get(m.group('name'), m.group('folder'), no_default=no_default) or ''
        return string


class _BaseFideliusAdminRepo(_BaseFideliusRepo, IFideliusAdminRepo, abc.ABC):
    """Covers a lot of admin basic functionality common across most storage back-ends.
    """
    def __init__(self, app_props: FideliusAppProps, tags: Optional[FideliusTags] = None, **kwargs):
        log.debug('_BaseFideliusAdminRepo.__init__ (this should set tags?!?)')
        super().__init__(app_props, **kwargs)
        self._tags = tags

    @property
    def tags(self) -> Optional[FideliusTags]:
        return self._tags

    def set_env(self, env: str):
        self.app_props.env = env



__all__ = [
    'IFideliusRepo',
    'IFideliusAdminRepo',
]
import abc
from fidelius.structs import *


class IFideliusRepo(abc.ABC):
    @abc.abstractmethod
    def __init__(self, app_props: FideliusAppProps, **kwargs):
        """Initialise a new Fidelius Repository.

        :param app_props: The application properties to use.
        """
        pass

    @property
    @abc.abstractmethod
    def app_props(self) -> FideliusAppProps:
        """The current application properties.
        """
        pass

    @abc.abstractmethod
    def make_app_path(self, env: Optional[str] = None) -> str:
        """The full path to application specific parameters/secrets.
        """
        pass

    @abc.abstractmethod
    def make_shared_path(self, folder: str, env: Optional[str] = None) -> str:
        """The full path to group shared parameters/secrets.
        """
        pass

    @abc.abstractmethod
    def get_expression_string(self, name: str, folder: Optional[str] = None) -> str:
        """Return a Fidelius expression string (e.g. to use in configuration
        files) which the `replace` method can parse and replace with
        parameters/secrets fetched from the storage backend.

        The default format of these expressions is:
        - `${__FID__:PARAM_NAME}` for app params/secrets
        - `${__FID__:FOLDER:PARAM_NAME}` for shared params/secrets in the given FOLDER

        :param name: The name of the parameter/secret to create an expression for.
        :param folder: Optional name of shared parameters/secrets to use. This
                       is optional and only applies to shared
                       parameters/secrets. Leaving it blank (default) will
                       return the app specific expression.
        :return: The Fidelius expression string.
        """
        pass

    @abc.abstractmethod
    def get_app_param(self, name: str, env: Optional[str] = None) -> Optional[str]:
        """Gets an application specific parameter/secret.

        :param name: The name of the parameter/secret to get.
        :param env: Optional env value override. Defaults to None, which uses
                    the env declaration from the current app properties.
        :return: The requested parameter/secret or None if it was not found.
        """
        pass

    @abc.abstractmethod
    def get(self, name: str, folder: Optional[str] = None, no_default: bool = False) -> Optional[str]:
        """Gets the given parameter/secret from the storage this repo uses.

        Parameters/secrets can be either application specific (use the
        `app_path`) or shared (use the `shared_path`) and this method should
        determine whether to get a shared or app parameter based on whether or
        not a folder name was given.

        Unless disabled by the `no_default` parameter, this method will attempt
        to fetch a parameter using the `env=default` if one was not found for
        the env in the current app properties.

        :param name: The name of the parameter/secret to get.
        :param folder: Optional name of shared parameters/secrets to use. This
                       is optional and only applies to shared
                       parameters/secrets. Leaving it blank (default) will
                       return the app specific parameter/secret.
        :param no_default: If True, does not try and get the default value if no
                           value was found for the current set environment.
        :return: The requested parameter/secret or None if it was not found.
        """
        pass

    @abc.abstractmethod
    def get_full_path(self, name: str, folder: Optional[str] = None, env: Optional[str] = None) -> str:
        """Gets the full path to a parameter/secret.

        Parameter/secret paths can be either application specific (use the
        `app_path`) or shared (use the `shared_path`) and this method should
        determine whether to use a shared or app path based on whether or
        not a folder name was given.

        :param name: The name of the parameter/secret to build a path to.
        :param folder: Optional name of shared parameters/secrets to use. This
                       is optional and only applies to shared
                       parameters/secrets. Leaving it blank (default) will
                       return the app specific parameter/secret path.
        :param env: Optional env value override. Defaults to None, which uses
                    the env declaration from the current app properties.
        :return: The full path to a parameter/secret.
        """
        pass

    @abc.abstractmethod
    def get_shared_param(self, name: str, folder: str, env: Optional[str] = None) -> Optional[str]:
        """Gets a group shared parameter/secret.

        :param name: The name of the parameter/secret to get.
        :param folder: The folder where the group shared parameter/secret is
                       located.
        :param env: Optional env value override. Defaults to None, which uses
                    the env declaration from the current app properties.
        :return: The requested parameter/secret or None if it was not found.
        """
        pass

    @abc.abstractmethod
    def replace(self, string: str, no_default: bool = False) -> str:
        """Take in a containing a Fidelius parameter/secret configuration
        expression (e.g. read from a config file), parses it, fetches the
        relevant parameter/secret and returns.

        The default format of these expressions is:
        - `${__FID__:PARAM_NAME}` for app params/secrets
        - `${__FID__:FOLDER:PARAM_NAME}` for shared params/secrets in the given FOLDER

        An empty string is returned if the parameter was not found and if the
        string does not match the expression format, it will be returned
        unchanged.

        :param string: The expression to replace with an actual parameter/secret
        :param no_default: If True, does not try and get the default value if no
                           value was found for the current set environment.
        :return: The requested value, an empty string or the original string
        """
        pass


class IFideliusAdminRepo(IFideliusRepo):
    @abc.abstractmethod
    def __init__(self, app_props: FideliusAppProps, tags: Optional[FideliusTags] = None, **kwargs):
        """Initialise a new Fidelius Admin Repository.

        :param app_props: The application properties to use.
        :param tags: An optional set of meta-data tags to use when creating new
                     parameters (if supported by the underlying
                     parameter/secret storage). Note that updating a parameter
                     does not update/change tags, they are only applied when
                     creating new parameters!
        """
        pass

    @property
    @abc.abstractmethod
    def tags(self) -> Optional[FideliusTags]:
        """The tags to use when creating new parameters.
        """
        pass

    @abc.abstractmethod
    def set_env(self, env: str):
        """Sets the current app properties env to something else (for admin purposes)
        """
        pass

    @abc.abstractmethod
    def create_param(self, name: str, value: str, description: Optional[str] = None, env: Optional[str] = None) -> (str, str):
        """Creates a new unencrypted application parameter.

        :param name: The parameter name.
        :param value: The parameter value.
        :param description: An optional description of the parameter (if
                            supported by the underlying parameter/secret
                            storage)
        :param env: Optional env value override. Defaults to None, which uses
                    the env from the current app properties.
        :return: A two tuple of strings containing the parameters full path and
                 the default suggested configuration expression to use (for
                 convenience, which the `replace` method can parse "as-is")
        """
        pass

    @abc.abstractmethod
    def update_param(self, name: str, value: str, description: Optional[str] = None, env: Optional[str] = None) -> (str, str):
        """Updates an existing unencrypted application parameter.

        :param name: The parameter name.
        :param value: The parameter value.
        :param description: An optional description of the parameter (if
                            supported by the underlying parameter/secret
                            storage)
        :param env: Optional env value override. Defaults to None, which uses
                    the env from the current app properties.
        :return: A two tuple of strings containing the parameters full path and
                 the default suggested configuration expression to use (for
                 convenience, which the `replace` method can parse "as-is")
        """
        pass

    @abc.abstractmethod
    def delete_param(self, name: str, env: Optional[str] = None):
        """Deletes an existing unencrypted application parameter.

        :param name: The parameter name.
        :param env: Optional env value override. Defaults to None, which uses
                    the env from the current app properties.
        """
        pass

    @abc.abstractmethod
    def create_shared_param(self, name: str, folder: str, value: str, description: Optional[str] = None, env: Optional[str] = None) -> (str, str):
        """Creates a new unencrypted group-shared parameter under the given
        shared folder.

        :param name: The parameter name.
        :param folder: The shared folder to place the parameter under.
        :param value: The parameter value.
        :param description: An optional description of the parameter (if
                            supported by the underlying parameter/secret
                            storage)
        :param env: Optional env value override. Defaults to None, which uses
                    the env from the current app properties.
        :return: A two tuple of strings containing the parameters full path and
                 the default suggested configuration expression to use (for
                 convenience, which the `replace` method can parse "as-is")
        """
        pass

    @abc.abstractmethod
    def update_shared_param(self, name: str, folder: str, value: str, description: Optional[str] = None, env: Optional[str] = None) -> (str, str):
        """Updates an existing unencrypted group-shared parameter under the
        given shared folder.

        :param name: The parameter name.
        :param folder: The shared folder to place the parameter under.
        :param value: The parameter value.
        :param description: An optional description of the parameter (if
                            supported by the underlying parameter/secret
                            storage)
        :param env: Optional env value override. Defaults to None, which uses
                    the env from the current app properties.
        :return: A two tuple of strings containing the parameters full path and
                 the default suggested configuration expression to use (for
                 convenience, which the `replace` method can parse "as-is")
        """
        pass

    @abc.abstractmethod
    def delete_shared_param(self, name: str, folder: str, env: Optional[str] = None):
        """Deletes an existing unencrypted group-shared parameter under the
        given shared folder.

        :param name: The parameter name.
        :param folder: The shared folder parameter is placed under.
        :param env: Optional env value override. Defaults to None, which uses
                    the env from the current app properties.
        """
        pass

    @abc.abstractmethod
    def create_secret(self, name: str, value: str, description: Optional[str] = None, env: Optional[str] = None) -> (str, str):
        """Creates a new encrypted application secret.

        :param name: The secret name.
        :param value: The secret value.
        :param description: An optional description of the secret (if
                            supported by the underlying parameter/secret
                            storage)
        :param env: Optional env value override. Defaults to None, which uses
                    the env from the current app properties.
        :return: A two tuple of strings containing the secret full path and
                 the default suggested configuration expression to use (for
                 convenience, which the `replace` method can parse "as-is")
        """
        pass

    @abc.abstractmethod
    def update_secret(self, name: str, value: str, description: Optional[str] = None, env: Optional[str] = None) -> (str, str):
        """Updates an existing encrypted application secret.

        :param name: The secret name.
        :param value: The secret value.
        :param description: An optional description of the secret (if
                            supported by the underlying parameter/secret
                            storage)
        :param env: Optional env value override. Defaults to None, which uses
                    the env from the current app properties.
        :return: A two tuple of strings containing the secret full path and
                 the default suggested configuration expression to use (for
                 convenience, which the `replace` method can parse "as-is")
        """
        pass

    @abc.abstractmethod
    def delete_secret(self, name: str, env: Optional[str] = None):
        """Deletes an existing encrypted application secret.

        :param name: The parameter name.
        :param env: Optional env value override. Defaults to None, which uses
                    the env from the current app properties.
        """
        pass

    @abc.abstractmethod
    def create_shared_secret(self, name: str, folder: str, value: str, description: Optional[str] = None, env: Optional[str] = None) -> (str, str):
        """Creates a new encrypted group-shared secret under the given
        shared folder.

        :param name: The secret name.
        :param folder: The shared folder to place the secret under.
        :param value: The secret value.
        :param description: An optional description of the parameter (if
                            supported by the underlying parameter/secret
                            storage)
        :param env: Optional env value override. Defaults to None, which uses
                    the env from the current app properties.
        :return: A two tuple of strings containing the secret full path and
                 the default suggested configuration expression to use (for
                 convenience, which the `replace` method can parse "as-is")
        """
        pass

    @abc.abstractmethod
    def update_shared_secret(self, name: str, folder: str, value: str, description: Optional[str] = None, env: Optional[str] = None) -> (str, str):
        """Updates an existing encrypted group-shared secret under the
        given shared folder.

        :param name: The secret name.
        :param folder: The shared folder to place the secret under.
        :param value: The secret value.
        :param description: An optional description of the parameter (if
                            supported by the underlying parameter/secret
                            storage)
        :param env: Optional env value override. Defaults to None, which uses
                    the env from the current app properties.
        :return: A two tuple of strings containing the secret full path and
                 the default suggested configuration expression to use (for
                 convenience, which the `replace` method can parse "as-is")
        """
        pass

    @abc.abstractmethod
    def delete_shared_secret(self, name: str, folder: str, env: Optional[str] = None):
        """Deletes an existing encrypted group-shared secret under the
        given shared folder.

        :param name: The secret name.
        :param folder: The shared folder secret is placed under.
        :param env: Optional env value override. Defaults to None, which uses
                    the env from the current app properties.
        """
        pass



"""Here we import this modules implementation classes with generic names for the
Factory to use"""
from ._mockrepo import MockFideliusRepo as FideliusRepo
from ._mockadmin import MockFideliusAdmin as FideliusAdmin



__all__ = [
    'MockFideliusRepo',
]

from fidelius.structs import *
from fidelius.gateway._abstract import *
from ._inmemcache import _SingletonDict

import base64

import logging
log = logging.getLogger(__name__)


class MockFideliusRepo(_BaseFideliusRepo):
    def __init__(self, app_props: FideliusAppProps, **kwargs):
        """This mock variation of the FideliusRepo simply returns a base64
        encoded version of the full path of the requested parameter/secret.

        This is mainly intended for unit testing other packages and apps that
        use Fidelius.
        """
        log.debug('MockFideliusRepo.__init__')
        super().__init__(app_props, **kwargs)
        self._cache: _SingletonDict[str, str] = _SingletonDict()

    def get_app_param(self, name: str, env: Optional[str] = None) -> Optional[str]:
        return self._cache.get(self.get_full_path(name, env=env), None)

    def get_shared_param(self, name: str, folder: str, env: Optional[str] = None) -> Optional[str]:
        return self._cache.get(self.get_full_path(name, folder, env=env), None)



__all__ = [
    'FideliusAppProps',
]
import dataclasses


@dataclasses.dataclass
class FideliusAppProps:
    """Application properties,  used to get or set specific variations of
    parameters/secrets.

    :ivar app: The name of the application itself (module name or slug)
    :ivar group: The app's "group" (think "business domain" or "namespace").
                 The purpose of the "group" attribute is to categorize shared
                 parameters/secrets such that all apps in the same group, can
                 access all the shared (non-app specific) parameters/secrets of
                 that group.
    :ivar env: Deployment environment (e.g., "prod", "dev", "test").
    """
    app: str
    group: str
    env: str = 'default'



__all__ = [
    'FideliusTags',
]
from ._base import *


class FideliusTags:
    __slots__ = ('application', 'owner', 'tier', 'finance', '_other')

    def __init__(self, application: str, owner: str, tier: str = 'default', finance: str = 'COST', **kwargs):
        self.application = application
        self.owner = owner
        self.tier = tier
        self.finance = finance
        self._other = kwargs or {}

    def __getattr__(self, name: str) -> Optional[str]:
        return self._other.get(name, None)

    def __setattr__(self, name: str, value: Optional[str]):
        if name in self.__slots__:
            if value is None:
                raise ValueError(f'tag "{name}" needs a string value')
            object.__setattr__(self, name, value)
        else:
            if value is None:
                if name in self._other:
                    del self._other[name]
            else:
                self._other[name] = value

    def __delattr__(self, name: str):
        self.__setattr__(name, None)

    def __repr__(self) -> str:
        tags = ', '.join([f"{k}='{v}'" for k, v in self.to_dict().items()])
        return f'{self.__class__.__name__}({tags})'

    def to_dict(self) -> Dict[str, str]:
        d = {
            'application': self.application,
            'owner': self.owner,
            'tier': self.tier,
            'finance': self.finance
        }
        if self._other:
            d.update(self._other)
        return d



__all__ = [
    'FideliusError',
    'FideliusAdminError',
    'FideliusParameterNotFound',
    'FideliusParameterAlreadyExists',
]


class FideliusError(Exception):
    pass


class FideliusAdminError(FideliusError):
    pass


class FideliusParameterNotFound(FideliusAdminError, KeyError):
    pass


class FideliusParameterAlreadyExists(FideliusAdminError, ValueError):
    pass



__all__ = [
    'FideliusFactory',
]

from fidelius.structs import *
from .interface import *
from ccptools.tpu import strimp

import logging
log = logging.getLogger(__name__)


class FideliusFactory:
    @staticmethod
    def get_class(impl: str = 'paramstore') -> Type[IFideliusRepo]:
        return strimp.get_class(f'fidelius.gateway.{impl}._std.FideliusRepo', logger=log, reraise=True)

    @staticmethod
    def get_admin_class(impl: str = 'paramstore') -> Type[IFideliusAdminRepo]:
        return strimp.get_class(f'fidelius.gateway.{impl}._std.FideliusAdmin', logger=log, reraise=True)



__all__ = [
    '_SingletonDict',
]

from ccptools.structs import *


class _SingletonDict(dict, metaclass=Singleton):
    """Simple Singleton dict :)

    It's point is simply to act as a shared centralized store for the mock
    stuff, mimicking how multiple instances of Fidelius Repos and/or Admin
    Repos would nevertheless fetch data from the same source.

    This is just to "mock" the shared parameter/secret stuff.

    ...sneaky, right?
    """
    pass



__all__ = [
    'AwsParameterStoreAdmin',
]

from fidelius.structs import *
from fidelius.gateway._abstract import *
from ._paramstorerepo import *

import logging
log = logging.getLogger(__name__)


class AwsParameterStoreAdmin(_BaseFideliusAdminRepo, AwsParamStoreRepo):
    def __init__(self, app_props: FideliusAppProps, tags: Optional[FideliusTags] = None, **kwargs):
        log.debug('AwsParameterStoreAdmin.__init__')
        super().__init__(app_props, tags, **kwargs)

    def create_param(self, name: str, value: str,
                     description: Optional[str] = None, env: Optional[str] = None) -> (str, str):
        path = self.get_full_path(name, env=env)
        res = self._set_parameter(full_name=path,
                                  value=value,
                                  description=description,
                                  overwrite=False,
                                  encrypted=False)
        return path, self.get_expression_string(name)

    def update_param(self, name: str, value: str,
                     description: Optional[str] = None, env: Optional[str] = None) -> (str, str):
        path = self.get_full_path(name, env=env)
        if self.get_app_param(name, env=env) is None:
            raise FideliusParameterNotFound(f'parameter not found: {path}')

        self._set_parameter(full_name=path,
                            value=value,
                            description=description,
                            overwrite=True,
                            encrypted=False)
        return path, self.get_expression_string(name)

    def delete_param(self, name: str, env: Optional[str] = None):
        self._delete_parameter(full_name=self.get_full_path(name, env=env))

    def create_shared_param(self, name: str, folder: str, value: str,
                            description: Optional[str] = None, env: Optional[str] = None) -> (str, str):
        path = self.get_full_path(name, folder=folder, env=env)
        self._set_parameter(full_name=path,
                            value=value,
                            description=description,
                            overwrite=False,
                            encrypted=False)
        return path, self.get_expression_string(name, folder=folder)

    def update_shared_param(self, name: str, folder: str, value: str,
                            description: Optional[str] = None, env: Optional[str] = None) -> (str, str):
        path = self.get_full_path(name, folder=folder, env=env)
        if self.get_shared_param(name, folder=folder, env=env) is None:
            raise FideliusParameterNotFound(f'parameter not found: {path}')

        self._set_parameter(full_name=path,
                            value=value,
                            description=description,
                            overwrite=True,
                            encrypted=False)
        return path, self.get_expression_string(name, folder=folder)

    def delete_shared_param(self, name: str, folder: str, env: Optional[str] = None):
        self._delete_parameter(full_name=self.get_full_path(name, folder=folder, env=env))

    def create_secret(self, name: str, value: str,
                      description: Optional[str] = None, env: Optional[str] = None) -> (str, str):
        path = self.get_full_path(name, env=env)
        self._set_parameter(full_name=path,
                            value=value,
                            description=description,
                            overwrite=False,
                            encrypted=True)
        return path, self.get_expression_string(name)

    def update_secret(self, name: str, value: str,
                      description: Optional[str] = None, env: Optional[str] = None) -> (str, str):
        path = self.get_full_path(name, env=env)
        if self.get_app_param(name, env=env) is None:
            raise FideliusParameterNotFound(f'parameter not found: {path}')

        self._set_parameter(full_name=path,
                            value=value,
                            description=description,
                            overwrite=True,
                            encrypted=True)
        return path, self.get_expression_string(name)

    def delete_secret(self, name: str, env: Optional[str] = None):
        self._delete_parameter(full_name=self.get_full_path(name, env=env))

    def create_shared_secret(self, name: str, folder: str, value: str,
                             description: Optional[str] = None, env: Optional[str] = None) -> (str, str):
        path = self.get_full_path(name, folder=folder, env=env)
        self._set_parameter(full_name=path,
                            value=value,
                            description=description,
                            overwrite=False,
                            encrypted=True)
        return path, self.get_expression_string(name, folder=folder)

    def update_shared_secret(self, name: str, folder: str, value: str,
                             description: Optional[str] = None, env: Optional[str] = None) -> (str, str):
        path = self.get_full_path(name, folder=folder, env=env)
        if self.get_shared_param(name, folder=folder, env=env) is None:
            raise FideliusParameterNotFound(f'parameter not found: {path}')

        self._set_parameter(full_name=path,
                            value=value,
                            description=description,
                            overwrite=True,
                            encrypted=True)
        return path, self.get_expression_string(name, folder=folder)

    def delete_shared_secret(self, name: str, folder: str, env: Optional[str] = None):
        self._delete_parameter(full_name=self.get_full_path(name, folder=folder, env=env))

    def _tags_to_aws_format(self) -> Optional[List[Dict[str, str]]]:
        if self.tags:
            return [{'Key': k, 'Value': v} for k, v in self.tags.to_dict().items()]
        return None

    def _set_parameter(self,
                       full_name: str,
                       value: str,
                       encrypted: bool = False,
                       overwrite: bool = False,
                       description: Optional[str] = None) -> Dict:
        kwargs = dict(Name=full_name,
                      Description=description or full_name,
                      Value=value,
                      Type='SecureString' if encrypted else 'String',
                      Overwrite=overwrite,
                      Tier='Standard')
        if not overwrite:
            tags = self._tags_to_aws_format()
            if tags:
                kwargs['Tags'] = tags

        if encrypted:
            kwargs['KeyId'] = self._aws_key_arn

        try:
            response = self._ssm.put_parameter(**kwargs)
            return response
        except self._ssm.exceptions.ParameterAlreadyExists:
            raise FideliusParameterAlreadyExists(f'parameter already exists: {full_name}')

        except self._ssm.exceptions.ParameterNotFound:
            raise FideliusParameterNotFound(f'parameter not found: {full_name}')

    def _delete_parameter(self, full_name: str) -> Dict:
        try:
            response = self._ssm.delete_parameter(Name=full_name)
            return response
        except self._ssm.exceptions.ParameterNotFound:
            raise FideliusParameterNotFound(f'parameter not found: {full_name}')



__all__ = [
    'MockFideliusAdmin',
]

from fidelius.gateway._abstract import *
from fidelius.structs import *
from ._mockrepo import *

import logging
log = logging.getLogger(__name__)


class MockFideliusAdmin(_BaseFideliusAdminRepo, MockFideliusRepo):
    def __init__(self, app_props: FideliusAppProps, tags: Optional[FideliusTags] = None, **kwargs):
        """This mock version of the Fidelius Admin stores created and updated
        params in memory only (although the cache is a singleton so multiple
        instances of both admin and repo will be useing the same dict/data.

        Note that it does NOT extend the functionality of its non-Admin sibling,
        the MockFideliusRepo and thus does not return a base64 encoded version
        of every requested param/secret key name, but instead only uses its own
        internal in-memory cache, and thus, `get` will not return anything that
        hasn't been created first during that particular runtime.

        This is mainly intended for unit testing other packages and apps that
        use Fidelius.
        """
        log.debug('MockFideliusAdmin.__init__')
        super().__init__(app_props, tags, **kwargs)

    def _create(self, name: str, value: str, env: Optional[str] = None, folder: Optional[str] = None) -> (str, str):
        key = self.get_full_path(name, folder=folder, env=env)
        if key in self._cache:
            raise FideliusParameterAlreadyExists(f'parameter already exists: {key}')
        self._cache[key] = value
        return key, self.get_expression_string(name, folder=folder)

    def _update(self, name: str, value: str, env: Optional[str] = None, folder: Optional[str] = None) -> (str, str):
        key = self.get_full_path(name, folder=folder, env=env)
        if key not in self._cache:
            raise FideliusParameterNotFound(f'parameter not found: {key}')
        self._cache[key] = value
        return key, self.get_expression_string(name, folder=folder)

    def _delete(self, name: str, env: Optional[str] = None, folder: Optional[str] = None):
        key = self.get_full_path(name, folder=folder, env=env)
        if key not in self._cache:
            raise FideliusParameterNotFound(f'parameter not found: {key}')
        del self._cache[key]

    def create_param(self, name: str, value: str,
                     description: Optional[str] = None, env: Optional[str] = None) -> (str, str):
        return self._create(name, value=value, env=env)

    def update_param(self, name: str, value: str,
                     description: Optional[str] = None, env: Optional[str] = None) -> (str, str):
        return self._update(name, value=value, env=env)

    def delete_param(self, name: str, env: Optional[str] = None):
        self._delete(name, env=env)

    def create_shared_param(self, name: str, folder: str, value: str,
                            description: Optional[str] = None,
                            env: Optional[str] = None) -> (str, str):
        return self._create(name, value=value, env=env, folder=folder)

    def update_shared_param(self, name: str, folder: str, value: str,
                            description: Optional[str] = None,
                            env: Optional[str] = None) -> (str, str):
        return self._update(name, value=value, env=env, folder=folder)

    def delete_shared_param(self, name: str, folder: str, env: Optional[str] = None):
        self._delete(name, env=env, folder=folder)

    def create_secret(self, name: str, value: str,
                      description: Optional[str] = None, env: Optional[str] = None) -> (str, str):
        return self._create(name, value=value, env=env)

    def update_secret(self, name: str, value: str,
                      description: Optional[str] = None, env: Optional[str] = None) -> (str, str):
        return self._update(name, value=value, env=env)

    def delete_secret(self, name: str, env: Optional[str] = None):
        self._delete(name, env=env)

    def create_shared_secret(self, name: str, folder: str, value: str,
                             description: Optional[str] = None, env: Optional[str] = None) -> (str, str):
        return self._create(name, value=value, env=env, folder=folder)

    def update_shared_secret(self, name: str, folder: str, value: str,
                             description: Optional[str] = None, env: Optional[str] = None) -> (str, str):
        return self._update(name, value=value, env=env, folder=folder)

    def delete_shared_secret(self, name: str, folder: str, env: Optional[str] = None):
        self._delete(name, env=env, folder=folder)



__all__ = [
    'AwsParamStoreRepo',
]

from fidelius.structs import *
from fidelius.gateway._abstract import *


import boto3
import os

import logging
log = logging.getLogger(__name__)


class AwsParamStoreRepo(_BaseFideliusRepo):
    def __init__(self, app_props: FideliusAppProps,
                 aws_access_key_id: str = None,
                 aws_secret_access_key: str = None,
                 aws_key_arn: str = None,
                 aws_region_name: str = None,
                 aws_endpoint_url: str = None,
                 flush_cache_every_time: bool = False,
                 **kwargs):
        """Fidelius Admin Repo that uses AWS' Simple Systems Manager's Parameter Store as a back end.

        :param app_props: The current application properties.
        :param aws_access_key_id: Optional AWS_ACCESS_KEY_ID which is otherwise
                                  pulled from the FIDELIUS_AWS_ACCESS_KEY_ID or
                                  AWS_ACCESS_KEY_ID environment variables.
        :param aws_secret_access_key: Optional AWS_SECRET_ACCESS_KEY which is otherwise
                                      pulled from the FIDELIUS_AWS_SECRET_ACCESS_KEY or
                                      AWS_SECRET_ACCESS_KEY environment variables.
        :param aws_key_arn: Optional ARN to an AWS KMS encryption key that'll be
                            used to encrypt secret parameters. If not provided
                            it'll be extracted from the FIDELIUS_AWS_KEY_ARN
                            environment variable and if that is missing as well,
                            an EnvironmentError is raised.
        :param aws_region_name: Optional AWS region name, which is otherwise
                                extracted from the FIDELIUS_AWS_REGION_NAME or
                                AWS_DEFAULT_REGION environment variables or just
                                set to `eu-west-1` by default is completely
                                missing.
        :param aws_endpoint_url: Optional custom AWS endpoint URL intended for
                                 testing and development, e.g. by spinning up a
                                 LocalStack container and pointing to that
                                 instead of a live AWS environment.
        :param flush_cache_every_time: Optional flat that'll flush the entire
                                       cache before every operation if set to
                                       True and is just intended for testing
                                       purposes.
        """
        super().__init__(app_props, **kwargs)
        self._flush_cache_every_time = flush_cache_every_time

        self._aws_key_arn = aws_key_arn or os.environ.get('FIDELIUS_AWS_KEY_ARN', '')
        if not self._aws_key_arn:
            raise EnvironmentError('Fidelius AwsParamStoreRepo requires the ARN for the KMS key argument when initialising or in the FIDELIUS_AWS_KEY_ARN environment variable')

        self._region_name = aws_region_name or os.environ.get('FIDELIUS_AWS_REGION_NAME', None) or os.environ.get('AWS_DEFAULT_REGION', 'eu-west-1')

        self._aws_endpoint_url = aws_endpoint_url or os.environ.get('FIDELIUS_AWS_ENDPOINT_URL', '')

        self._force_log_secrecy()
        self._ssm = boto3.client('ssm',
                                 region_name=self._region_name,
                                 endpoint_url=self._aws_endpoint_url or None,
                                 aws_access_key_id=aws_access_key_id or os.environ.get('FIDELIUS_AWS_ACCESS_KEY_ID', None) or os.environ.get('AWS_ACCESS_KEY_ID', None),
                                 aws_secret_access_key=aws_secret_access_key or os.environ.get('FIDELIUS_AWS_SECRET_ACCESS_KEY', None) or os.environ.get('AWS_SECRET_ACCESS_KEY', None))

        self._cache: Dict[str, str] = {}
        self._loaded_paths: Set[str] = set()

    def _nameless_path(self, folder: Optional[str] = None, env: Optional[str] = None) -> str:
        return self.get_full_path(name='', folder=folder, env=env)[:-1]

    @staticmethod
    def _force_log_secrecy():
        # We don't allow debug or less logging of botocore's HTTP requests cause
        # those logs have unencrypted passwords in them!
        botolog = logging.getLogger('botocore')
        if botolog.level < logging.INFO:
            botolog.setLevel(logging.INFO)

    def _load_path(self, folder: Optional[str] = None, env: Optional[str] = None):
        log.debug('AwsParamStoreRepo._load_path(folder=%s, env=%s)', folder, env)
        self._force_log_secrecy()

        # This is stuff for unit-testing only!
        if self._flush_cache_every_time:
            self._loaded_paths = set()
            self._cache = {}

        path = self._nameless_path(folder, env)
        if path in self._loaded_paths:
            return

        response = self._ssm.get_parameters_by_path(
            Path=path,
            Recursive=True,
            WithDecryption=True
        )

        for p in response.get('Parameters', []):
            key = p.get('Name')
            if key:
                self._cache[key] = p.get('Value')

        self._loaded_paths.add(path)

    def get_app_param(self, name: str, env: Optional[str] = None) -> Optional[str]:
        self._load_path(env=env)
        return self._cache.get(self.get_full_path(name, env=env), None)

    def get_shared_param(self, name: str, folder: str, env: Optional[str] = None) -> Optional[str]:
        self._load_path(folder=folder, env=env)
        return self._cache.get(self.get_full_path(name, folder, env=env), None)
